#! /usr/bin/perl -w

%non_aggregate = ( 
    GtkType => 1, GdkAtom => 1, GtkFunction => 1,
    GtkMatchType => 1, GtkCellType => 1, GtkToolbarChildType => 1);

%redeclare = ( "gtk_object" => "gtk_signal" );    
%redeclare_source = ( "gtk_signal" => "gtksignal.h" );

%strunames = ();
%functions = ();

%opt = ();
%shortopt = ( "-v" => "verbose", "-q" => "quiet", "-m" => "methods", "-fk" => "functions");
for (values %shortopt) { $opt{$_} = 0; }

$addE = "";

for $F (@ARGV)
{
  if (exists $shortopt{$F}) { $opt{$shortopt{$F}} = 1; next; }
  if ($F =~ m/^--(.*)$/) { $opt{$1} = 1; next; }
  open F, "<$F" or next; $T = join( "",<F>); close F;
  
  %strunames = ();
  %functions = ();
  $addE = "";
  
  $T =~ s/(^)(\s*\#\s*define.*)$/$1$2###/mg;

  $T =~ s/ (struct\s+)(\w+) (\s+{) 
      	    (\s+ (?:\/\* [^\/]+ \*\/ \s+)? )
            (\w+)([\s\*]+)(\w+)(\;) 
         / &register_structhead($1,$2,$3,$4,$5,$6,$7,$8) /gsex;
         
  $T =~ s/\<gtk\//\<gtk\+\//g;
  
  $T =~ s/([\w\s\*]+) (\s)(\w+) (\s*\(\s*) ([\w\s\*\,]+) (\s*\)\s*) (\;)
         / &register_function($1,$2,$3,$4,$5,$6,$7) /gsex;
         
  for my $i (keys %strunames) { 
      if (exists $redeclare{$i}) {
          redeclare_functions($i);
          my $f = $redeclare_source{$redeclare{$i}};
          if (defined $f) {
              $T =~ s/ (\#ifdef \s+__cplusplus \s+extern \s+\"C\")
                     / "\n\#include <gtk+\/$f>\n$1" /sex;
          }
      }
  }
         
  print $F,": " unless $opt{"quiet"};        
  print join (" ", keys %strunames) unless $opt{"quiet"};
  print " ... " unless $opt{"quiet"};
  print join (" ", keys %functions) if $opt{"verbose"};
  
  for my $stru (keys %strunames)
  {
      my $struname = $strunames{$stru};
      
      $T =~ s/ (struct\s+ $struname \s* 
               (?:__gtk_inherits\s*\(\s*\w+\s*\)\s*)?
               \{ [^\}]+) (\}\s*\;)
             / &add_inlines($stru,$1,$2) /gsex;
  }
  
  if ($F eq "gtk.h") { 
      	$T = 
         "\n#ifdef __cplusplus"
        ."\n#define __gtk_inherits(x) : x"
        ."\n#define __GTK_inherits(x,y)"
        ."\n#else"
        ."\n#define __gtk_inherits(x)"
        ."\n#define __GTK_inherits(x,y)  x y ;"
        ."\n#endif"
        ."\n".$T
        ."\n#define GTK_SIG (void (*)(...))\n"; }
        
  $T =~ s/ (\#ifdef\s*__cplusplus\s*}) (\s*\#endif) 
         /$1." \/\* end of C linkage \*\/\n\n".$addE.$2/sex;

  $T =~ s/###$//mg;

  print "\n" unless $opt{"quiet"};
  open E, ">../gtk+/$F" or next; print E $T; close E;
}

sub register_structhead
{
  my ($struP,$stru,$brac,$bracW,$base,$baseW,$name,$nameS) = @_;
        
  $struname = $stru;
  $struname =~ s/ ^_?([A-Z]) / lc($1) /ex;
  $struname =~ s/ ([A-Z]+) / "_".lc($1) /gex;
  $struname =~ s/^_+//;
  if ($struname =~ m/_/) 
  {   $strunames{"$struname"} = $stru;
  }


  if (exists $non_aggregate{"$base"} or $baseW !~ m/^\s+$/
        or $base =~ m/^[a-z]+[1-9]*$/)
        { return $struP.$stru.$brac.$bracW.$base.$baseW.$name.$nameS; }


  return $struP.$stru
       ."\n  __gtk_inherits( ".$base." ) "
       .$brac." __GTK_inherits( ".$base.", ".$name." )".$bracW ;
}

sub register_function
{
   my ($rtype,$nameP,$name,$lbrc,$args,$rbrc,$semi) = @_;
   my $line = join ("",@_);
   
   print "f:$name " if $opt{"functions"};
   
   $functions{"$name"} = $rtype.$nameP.$name.$lbrc.$args.$rbrc;

   return $line;
}

sub add_inlines
{
   my ($stru,$pre,$post) = @_;
   my $add = "";

   print "  [INLINES for $stru: " if $opt{"verbose"};
   
   for $f (sort keys %functions)
   {
       if ($f =~ m/^${stru}_(.*)/ 
       or defined $redeclare{$stru} and $f =~ m/^$redeclare{$stru}_(.*)/)
       {
           print $1," " if $opt{"verbose"};
           $shortname = $1;
           $longname = $f;
           $struct_name = "struct $strunames{$stru}";
           $funcdef  = $functions{$f};
           $rettype  = $funcdef;
           $rettype  =~ s/ ^ ([^\(]+) \(.* / $1 /sx;
           $rettype =~ s/\s+/ /gs;
           $rettype  =~ s/\s$//; $rettype =~ s/\w+$//;
           $rettype  =~ s/^\s//; $rettype =~ s/\s+$//;   
           $funcdecl = $funcdef;
           $funcdecl =~ s/ ^ [^\(]+ \( / \( /x;
           $funcargs = $funcdecl;
           $funcargs =~ s/ ([\w\s\*]+) (\s|\*)(\w+\s*) ([\)\,]) / $3$4 /gsx;
           $funcargs =~ s/void//g;
           $funcdecl =~ s/\s+/ /g; $funcdecl =~ s/^\s//;  $funcdecl =~ s/\s$//;
           $funcargs =~ s/\s+/ /g; $funcargs  =~ s/^\s//; $funcargs  =~ s/\s$//;
           
           $firsttype = $funcdecl;
           $firsttype =~ s/\s*\(\s*(\w+)(\s|\*).*/$1/;
           
           # the GTK people are severely misguided on returning Widget from any new-function
           $retconv = "";
           if ($shortname =~ m/^new$/ or $shortname =~ m/^new_/) {
               $rettype = "$strunames{$stru}*";
               $retconv = "($rettype)";
           }


               $add .= "  inline static $rettype _${shortname} $funcdecl; \n";
           if ($rettype eq "void")
           {
               $addE .= "inline $rettype\n"
                       ."$strunames{$stru}::_${shortname} $funcdecl \n"
                       ."                { $longname$funcargs; } \n";
           }else{
               $addE .= "inline $rettype\n"
                       ."$strunames{$stru}::_${shortname} $funcdecl \n"
                       ."                { return $retconv$longname$funcargs; } \n";
           }
             
           if ($strunames{$stru} eq "_$firsttype" || $struct_name eq $firsttype)
           {
                $funcdecl =~ s/ (\() [^\,\)]+ ([\,\)]) /$1.($2 eq ")"?")":"")/sex;
                $funcargs =~ s/ (\() [^\,\)]+ ([\,\)]) /$1."this".$2/sex;
                if ($rettype eq "void")
                {
                   # compute the type and name of the second(!) argument
                   if ( $funcargs =~ m/,/)
                   {
                      $funcdecl =~ m/ \s*\( ([^\,\)]+) [\,\)] /sx;
                      $arg2 = $1; $arg2 =~ m/^(.*[\s\*])(\w+)\s*$/s;
                      $rettype = $1; $retconv = "($2)";
                $addE .= "  inline $rettype\n"
                        ."$strunames{$stru}::${shortname}_ $funcdecl \n"
                        ."                { $longname$funcargs; return $retconv; } \n";
                   }else{

                $addE .= "inline $rettype\n"
                        ."$strunames{$stru}::${shortname}_ $funcdecl \n"
                        ."                { $longname$funcargs; } \n";
                   }
                }else{
                $addE .= "  inline $rettype\n"
                        ."$strunames{$stru}::${shortname}_ $funcdecl \n"
                        ."                { return $retconv$longname$funcargs; } \n";
                }
                $add .= "  inline        $rettype  ${shortname}_ $funcdecl; \n";
                print $shortname." " if $opt{"methods"};
           }
       }
   }
   print "]\n" if $opt{"verbose"};
   
   if (length $add > 1)
   { 
       return  $pre."\n# if defined __cplusplus || defined __cplus\n"
              .$add."\n# endif /* __cplus(plus) */\n"
              .$post;
   }else{
       return $pre.$add.$post;
   }
}

sub redeclare_functions
{
  my $stru = $_[0];

  print "$stru -> $redeclare{$stru} -> $redeclare_source{$redeclare{$stru}}";
  return if not exists $redeclare{$stru} or not exists $redeclare_source{$redeclare{$stru}};
  
  my $R = $redeclare_source{$redeclare{$stru}};

  my $S = "";
  open R,"<$R" or return; $S = join("",<R>); close R;
  
  $S =~
        s/([\w\s\*]+) (\s) (\w+) (\s*\(\s*) ([\w\s\*\,]+) (\s*\)\s*) (\;)
         / &register_function($1,$2,$3,$4,$5,$6,$7) /gsex;  

  return;
}

